# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import inspect
import configparser
from os.path import join
from .conftest import SourceTree


def test_entrypoint():    
    
    setup_cfg = configparser.ConfigParser()
    setup_cfg.read( join(SourceTree,'setup.cfg') )
    console_scripts = setup_cfg['options.entry_points']['console_scripts']
    
    entry_point = console_scripts.split('=')[-1].strip().split(':')
    module_path = entry_point[0]
    method_name = entry_point[1]
    
    namespace = {}
    exec("from {} import {}".format(module_path, method_name), namespace)
    
    assert method_name in namespace
    function = namespace[method_name]
    
    assert callable(function)
    print( inspect.signature(function) )



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import ast
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parser import ArgParser
from os.path import (
    join,
    abspath,
    basename,
    splitext,
)


def rotation_type(string):
    rotation = int(string)
    if rotation not in (0, 90, 180, 270):
        raise ValueError("Invalid rotation value {}".format(rotation))
    return rotation


def colour_type(string):
    
    if string.lower() == 'none':
        return None
        
    else:
        
        colour = ast.literal_eval(string)
        
        if not isinstance(colour, (tuple, list)):
            raise ValueError( "Invalid colour type {}. Must be list or tuple.".format(type(colour)) )
        if not len(colour) in (3, 4):
            raise ValueError("Invalid number of colour values. Must be 3 or 4.")
        if not all(isinstance(val, int) and 0 <= val <= 255 for val in colour):
            raise ValueError("Colour values must be integers ranging from 0 to 255.")
        
        return colour


def pagetext_type(value):
    
    if not value:
        return
    
    page_indices = []
    splitted = value.split(',')
    
    for page_or_range in splitted:
        
        if '-' in page_or_range:
            
            start, end = page_or_range.split('-')
            start = int(start) - 1
            end = int(end) - 1
            
            if start < end:
                pages = [i for i in range(start, end+1)]
            else:
                pages = [i for i in range(start, end-1, -1)]
            
            page_indices.extend(pages)
        
        else:
            
            page_indices.append(int(page_or_range) - 1)
    
    return page_indices


def parse_args(argv, prog, desc):
    
    parser = ArgParser(
        prog = prog,
        description = desc,
    )
    
    parser.add_argument(
        'inputs',
        nargs = '+',
        help = "PDF documents to render",
    )
    parser.add_argument(
        '--output', '-o',
        type = abspath,
        required = True,
        help = "Output directory where to place the serially numbered images",
    )
    parser.add_argument(
        '--format', '-f',
        default = 'png',
        help = "File extension of the image format to use",
    )
    parser.add_argument(
        '--pages',
        default = None,
        type = pagetext_type,
        help = "Numbers of the pages to render (defaults to all)",
    )
    parser.add_argument(
        '--scale',
        default = 1,
        type = float,
        help = "Define the resolution of the output images. By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the number of pixels that represent one point.",
    )
    parser.add_argument(
        '--rotation',
        default = 0,
        type = rotation_type,
        help = "Rotate pages by 90, 180 or 270 degrees",
    )
    parser.add_argument(
        '--colour',
        default = (255, 255, 255, 255),
        type = colour_type,
        help = "Page background colour. Defaults to white. It can be given in RGBA format as a sequence of integers ranging from 0 to 255, or it may be 'none' for transparent background."
    )
    parser.add_argument(
        '--no-annotations',
        action = 'store_true',
        help = "Option to prevent rendering of PDF annotations",
    )
    parser.add_argument(
        '--optimise-mode',
        default = pdfium.OptimiseMode.none,
        type = lambda string: pdfium.OptimiseMode[string.lower()],
        help = "Select a rendering optimisation mode (none, lcd_display, printing)",
    )
    parser.add_argument(
        '--greyscale',
        action = 'store_true',
        help = "Whether to render in greyscale mode (no colours)",
    )
    parser.add_argument(
        '--processes',
        default = os.cpu_count(),
        type = int,
        help = "The number of processes to use for rendering (defaults to the number of CPU cores)"
    )
    
    return parser.parse_args(argv)


def main(argv, prog, desc):
    
    args = parse_args(argv, prog, desc)
    
    for input_path in args.inputs:
        
        prefix = splitext(basename(input_path))[0] + '_'
        
        renderer = pdfium.render_pdf_topil(
            input_path,
            page_indices = args.pages,
            scale = args.scale,
            rotation = args.rotation,
            colour = args.colour,
            annotations = not args.no_annotations,
            greyscale = args.greyscale,
            optimise_mode = args.optimise_mode,
            n_processes = args.processes,
        )
        
        for image, suffix in renderer:
            output_path = "{}.{}".format(join(args.output, prefix+suffix), args.format)
            image.save(output_path)
            image.close()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Attempt to download and build PDFium from source. This may take very long.
# Only tested on Linux with glibc. Might not work on other platforms.

import os
import sys
import shutil
import argparse
import importlib.util
from os.path import join


if __name__ == '__main__': sys.modules['platform_setup'] = importlib.util.module_from_spec( importlib.util.spec_from_file_location('platform_setup', join(os.path.dirname(os.path.abspath(__file__)), '__init__.py')) )

from platform_setup import check_deps
from platform_setup.packaging_base import (
    SB_Dir,
    Libnames,
    DataTree,
    PlatformNames,
    run_cmd,
    call_ctypesgen,
)


PatchDir       = join(SB_Dir,'patches')
DepotToolsDir  = join(SB_Dir,'depot_tools')
PDFiumDir      = join(SB_Dir,'pdfium')
PDFiumBuildDir = join(PDFiumDir,'out','Default')
OutputDir      = join(DataTree,PlatformNames.sourcebuild)
NB_BinaryDir   = join(PDFiumDir,'third_party','llvm-build','Release+Asserts','bin')

DepotTools_URL = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
PDFium_URL     = "https://pdfium.googlesource.com/pdfium.git"

DepotPatches = [
    (join(PatchDir,'depot_tools','gclient_scm.patch'), DepotToolsDir),
]
PdfiumMainPatches = [
    (join(PatchDir,'pdfium','public_headers.patch'), PDFiumDir),
    (join(PatchDir,'pdfium','shared_library.patch'), PDFiumDir),
]
PdfiumWinPatches = [
    (join(PatchDir,'pdfium','win','pdfium.patch'), PDFiumDir),
    (join(PatchDir,'pdfium','win','build.patch'), join(PDFiumDir,'build')),
]
PdfiumNativebuildPatches = [
    (join(PatchDir,'pdfium','nativebuild.patch'), join(PDFiumDir,'build')),
]

DefaultConfig = {
    'is_debug': False,
    'treat_warnings_as_errors': False,
    'pdf_is_standalone': True,
    'pdf_enable_v8': False,
    'pdf_enable_xfa': False,
    'pdf_use_skia': False,
}
NativebuildConfig = {
    'clang_use_chrome_plugins': False,
    'init_stack_vars': False,
    'use_cxx11': True,
}
SyslibsConfig = {
    'use_system_freetype': True,
    'use_system_lcms2': True,
    #'use_system_libjpeg': True,
    'use_system_libopenjpeg2': True,
    'use_system_libpng': True,
    'use_system_zlib': True,
    'sysroot': '/',
}

if sys.platform.startswith('linux'):
    DefaultConfig['use_custom_libcxx'] = True
    SyslibsConfig['use_custom_libcxx'] = False
elif sys.platform.startswith('darwin'):
    DefaultConfig['mac_deployment_target'] = '10.11.0'
    SyslibsConfig['use_system_xcode'] = True
elif sys.platform.startswith('win32'):
    DefaultConfig['pdf_use_win32_gdi'] = True


def dl_depottools(do_update):
    
    if not os.path.isdir(SB_Dir):
        os.makedirs(SB_Dir)
    
    is_update = True
    
    if os.path.isdir(DepotToolsDir):
        if do_update:
            print("DepotTools: Revert and update ...")
            run_cmd(['git', 'reset', '--hard', 'HEAD'], cwd=DepotToolsDir)
            run_cmd(['git', 'pull', DepotTools_URL], cwd=DepotToolsDir)
        else:
            print("DepotTools: Using existing repository as-is.")
            is_update = False
    else:
        print("DepotTools: Download ...")
        run_cmd(['git', 'clone', '--depth', '1', DepotTools_URL, DepotToolsDir], cwd=SB_Dir)
    
    os.environ['PATH'] += os.pathsep + DepotToolsDir
    
    return is_update


def _apply_patchset(patchset):
    for patch, cwd in patchset:
        run_cmd(['git', 'apply', '-v', patch], cwd=cwd)


def dl_pdfium(do_update, revision, GClient):
    
    is_sync = True
    
    if os.path.isdir(PDFiumDir):
        if do_update:
            print("PDFium: Revert / Sync  ...")
            run_cmd([GClient, 'revert'], cwd=SB_Dir)
        else:
            is_sync = False
            print("PDFium: Using existing repository as-is.")
    else:
        print("PDFium: Download ...")
        run_cmd([GClient, 'config', '--unmanaged', PDFium_URL], cwd=SB_Dir)
    
    if is_sync:
        run_cmd([GClient, 'sync', '--revision', 'origin/{}'.format(revision), '--no-history', '--shallow'], cwd=SB_Dir)
    
    return is_sync


def patch_depottools():
    _apply_patchset(DepotPatches)


def _find_latest_llvm():
    
    # assuming there is a corresponding lld installed for the latest llvm
    
    llvm_prefix = 'llvm-'
    latest = None
    libdir = None
    
    for search_dir in ('/usr/lib', '/usr/local/lib'):
        for entry in os.listdir(search_dir):
            if not os.path.isdir( join(search_dir, entry) ):
                continue
            if not entry.startswith('llvm-'):
                continue
            version = int(entry.split('-')[-1])
            if (latest is None) or (version > latest):
                latest = version
                libdir = search_dir
    
    assert None not in (latest, libdir)
    latest = str(latest)
    
    return join(libdir, llvm_prefix+latest, 'bin')


def _replace_binaries():
    
    llvm_dir = _find_latest_llvm()
    not_found = []
    
    for name in os.listdir(NB_BinaryDir):
        
        binary_path = join(NB_BinaryDir, name)
        replacement = join(llvm_dir, name)
        os.remove(binary_path)
        
        if not os.path.isfile(replacement):
            not_found.append(name)
            continue
        
        os.symlink(replacement, binary_path)
    
    if len(not_found) > 0:
        print("Warning: No system-provided replacements available for {}".format(not_found), file=sys.stderr)


def patch_pdfium():
    
    _apply_patchset(PdfiumMainPatches)
    
    if sys.platform.startswith('win32'):
        _apply_patchset(PdfiumWinPatches)
        shutil.copy(join(PatchDir,'pdfium','win','resources.rc'), join(PDFiumDir,'resources.rc'))


def patch_pdfium_nativebuild():
    _apply_patchset(PdfiumNativebuildPatches)
    _replace_binaries()


def configure(config, GN):
    
    if not os.path.exists(PDFiumBuildDir):
        os.makedirs(PDFiumBuildDir)
    
    with open(join(PDFiumBuildDir,'args.gn'), 'w') as args_handle:
        args_handle.write(config)
    
    run_cmd([GN, 'gen', PDFiumBuildDir], cwd=PDFiumDir)


def build(Ninja):
    run_cmd([Ninja, '-C', PDFiumBuildDir, 'pdfium'], cwd=PDFiumDir)


def find_lib(srcname=None, directory=PDFiumBuildDir):
    
    if srcname is not None:
        path = join(PDFiumBuildDir, srcname)
        if os.path.isfile(path):
            return path
        else:
            print("Warning: The file of given srcname does not exist.", file=sys.stderr)
    
    libpath = None
    
    for lname in Libnames:
        path = join(directory, lname)
        if os.path.isfile(path):
            libpath = path
    
    if libpath is None:
        raise RuntimeError("Build artifact not found.")
    
    return libpath


def pack(src_libpath, destname=None):
    
    if os.path.isdir(OutputDir):
        shutil.rmtree(OutputDir)
    os.makedirs(OutputDir)
    
    if destname is None:
        destname = os.path.basename(src_libpath)
    
    destpath = join(OutputDir, destname)
    shutil.copy(src_libpath, destpath)
    
    include_dir = join(OutputDir,'include')
    shutil.copytree(join(PDFiumDir,'public'), include_dir)
    
    call_ctypesgen(OutputDir, include_dir)
    shutil.rmtree(include_dir)


def _get_tool(tool, win_append, is_nativebuild):
    
    exe = join(DepotToolsDir, tool)
    if sys.platform.startswith('win32'):
        exe += '.' + win_append
    
    if is_nativebuild:
        _sh_exe = shutil.which(tool)
        if _sh_exe:
            exe = _sh_exe
        else:
            print("Warning: Host system does not provide '{}'".format(tool), file=sys.stderr)
    
    return exe


def _serialise_config(config_dict):
    
    config_str = ''
    sep = ''
    
    for key, value in config_dict.items():
        config_str += sep + '{} = '.format(key)
        if isinstance(value, bool):
            config_str += str(value).lower()
        elif isinstance(value, str):
            config_str += '"{}"'.format(value)
        else:
            raise TypeError( "Not sure how to serialise type {}".format(type(value)) )
        sep = '\n'
    
    return config_str


def main(
        b_argfile = None,
        b_srcname = None,
        b_destname = None,
        b_update = False,
        b_checkdeps = False,
        b_nativebuild = False,
        b_use_syslibs = False,
        b_revision = None,
    ):
    
    if b_revision is None:
        b_revision = 'main'
    
    # on Linux, rename the binary to `pdfium` to ensure it also works with older versions of ctypesgen
    if b_destname is None and sys.platform.startswith('linux'):
        b_destname = 'pdfium'
    
    if b_nativebuild:
        print("Using binaries provided by the system, if available.")
    else:
        print("Using binaries provided by the PDFium toolchain.")
    
    if b_checkdeps:
        check_deps.main(b_nativebuild)
    
    GClient = _get_tool('gclient', 'bat', False)
    GN = _get_tool('gn', 'bat', b_nativebuild)
    Ninja = _get_tool('ninja', 'exe', b_nativebuild)
    
    if sys.platform.startswith('win32'):
        os.environ['DEPOT_TOOLS_WIN_TOOLCHAIN'] = "0"
    
    if b_argfile is None:
        config_dict = DefaultConfig.copy()
        if b_nativebuild:
            config_dict.update(NativebuildConfig)
        if b_use_syslibs:
            config_dict.update(SyslibsConfig)
        config_str = _serialise_config(config_dict)
        
    else:
        with open(os.path.abspath(b_argfile), 'r') as file_handle:
            config_str = file_handle.read()
    
    print("\nBuild configuration:\n{}\n".format(config_str))
    
    depot_dl_done = dl_depottools(b_update)
    if depot_dl_done:
        patch_depottools()
    
    pdfium_dl_done = dl_pdfium(b_update, b_revision, GClient)
    
    if pdfium_dl_done:
        patch_pdfium()
    if b_nativebuild:
        patch_pdfium_nativebuild()
    
    configure(config_str, GN)
    build(Ninja)
    
    libpath = find_lib(b_srcname)
    pack(libpath, b_destname)


def parse_args(argv):
    
    parser = argparse.ArgumentParser(
        description = "A script to automate building PDFium from source and generating bindings with ctypesgen.",
    )
    
    parser.add_argument(
        '--argfile', '-a',
        help = "A text file containing custom PDFium build configuration, to be evaluated by `gn gen`. Call `gn args --list out/Default/` in `sourcebuild/pdfium/` to obtain a list of possible options.",
    )
    parser.add_argument(
        '--srcname', '-s',
        help = "Name of the generated PDFium binary file. This script tries to automatically find the binary, which should usually work. If it does not, however, this option may be used to explicitly provide the file name to look for.",
    )
    parser.add_argument(
        '--destname', '-d',
        help = "Rename the binary to a different filename.",
    )
    parser.add_argument(
        '--update', '-u',
        action = 'store_true',
        help = "Update existing PDFium/DepotTools repositories, removing local changes.",
    )
    parser.add_argument(
        '--check-deps', '-c',
        action = 'store_true',
        help = "Check that all required dependencies are installed. (Automatically installs missing Python packages, complains about missing system dependencies.)",
    )
    parser.add_argument(
        '--nativebuild', '-n',
        action = 'store_true',
        help = "Try to use system-provided tools if available, rather than pre-built binaries from the PDFium toolchain. Warning: This may or may not work, and should only be used as last resort if the regular build strategy failed.",
    )
    parser.add_argument(
        '--use-syslibs', '-l',
        action = 'store_true',
        help = "Use system libraries instead of those bundled with PDFium. (Make sure that freetype, lcms2, libjpeg, libopenjpeg2, libpng and zlib are installed, and that $PKG_CONFIG_PATH is set correctly (e. g. to /usr/lib/x86_64-linux-gnu/pkgconfig)",
    )
    parser.add_argument(
        '--revision', '-r',
        help = "PDFium revision to check out (defaults to main).",
    )
    
    return parser.parse_args(argv)


def run_cli(argv=sys.argv[1:]):
    args = parse_args(argv)
    return main(
        b_argfile = args.argfile,
        b_srcname = args.srcname,
        b_destname = args.destname,
        b_update = args.update,
        b_checkdeps = args.check_deps,
        b_nativebuild = args.nativebuild,
        b_use_syslibs = args.use_syslibs,
        b_revision = args.revision,
    )
    

if __name__ == '__main__':
    run_cli()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

# Non-stdlib imports not allowed in this file, as it is imported prior to the check_deps call

import subprocess
from glob import glob
from os.path import (
    expanduser,
    dirname,
    abspath,
    join,
)

Libnames = (
    'pdfium',
    'pdfium.dylib',
    'pdfium.dll',
    'libpdfium.so',
    'pdfium.so',
    'libpdfium',
    'libpdfium.dylib',
    'libpdfium.dll',
)


HomeDir     = expanduser('~')
SourceTree  = dirname(dirname(abspath(__file__)))
DataTree    = join(SourceTree,'data')
SB_Dir      = join(SourceTree,'sourcebuild')
ModuleDir   = join(SourceTree,'src','pypdfium2')
VersionFile = join(ModuleDir,'_version.py')


class PlatformNames:
    darwin_x64    = 'darwin_x64'
    darwin_arm64  = 'darwin_arm64'
    linux_x64     = 'linux_x64'
    linux_x86     = 'linux_x86'
    linux_arm64   = 'linux_arm64'
    linux_arm32   = 'linux_arm32'
    windows_x64   = 'windows_x64'
    windows_x86   = 'windows_x86'
    windows_arm64 = 'windows_arm64'
    sourcebuild   = 'sourcebuild'


def run_cmd(command, cwd, **kwargs):
    print('{} ("{}")'.format(command, cwd))
    return subprocess.run(command, cwd=cwd, **kwargs)


def call_ctypesgen(platform_dir, include_dir):
    
    bindings_file = join(platform_dir,'_pypdfium.py')
    
    ctypesgen_cmd = ['ctypesgen', '--library', 'pdfium', '--strip-build-path', platform_dir, '-L', '.'] + sorted(glob( join(include_dir,'*.h') )) + ['-o', bindings_file]
    run_cmd(ctypesgen_cmd, cwd=platform_dir)
    
    with open(bindings_file, 'r') as file_reader:
        text = file_reader.read()
        text = text.replace(platform_dir, '.')
        text = text.replace(HomeDir, '~')
    
    with open(bindings_file, 'w') as file_writer:
        file_writer.write(text)


def _get_ver_namespace():
    
    ver_namespace = {}
    with open(VersionFile, 'r') as fh:
        exec(fh.read(), ver_namespace)
    
    return ver_namespace


_ver_namespace = _get_ver_namespace()

def extract_version(variable_str):
    return _ver_namespace[variable_str]



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import logging
import os
from os.path import (
    join,
    dirname,
    abspath,
)


lib_logger = logging.getLogger('pypdfium2')
lib_logger.addHandler(logging.StreamHandler())


TestDir     = dirname(abspath(__file__))
SourceTree  = dirname(TestDir)
ResourceDir = join(TestDir,'resources')
OutputDir   = join(TestDir,'output')


class TestFiles:
    render = join(ResourceDir,'render.pdf')
    encrypted = join(ResourceDir,'encrypted.pdf')
    multipage = join(ResourceDir,'multipage.pdf')
    bookmarks = join(ResourceDir,'bookmarks.pdf')
    bookmarks_circular = join(ResourceDir,'bookmarks_circular.pdf')
    cropbox = join(ResourceDir,'cropbox.pdf')
    mediabox_missing = join(ResourceDir,'mediabox_missing.pdf')
    nonascii = join(ResourceDir,'nonascii_tênfilechứakýtựéèáàçß 发短信.pdf')


def iterate_testfiles(skip_encrypted=True):
    
    encrypted = (TestFiles.encrypted, )
    
    for attr_name in dir(TestFiles):
        
        if attr_name.startswith('_'):
            continue
        
        member = getattr(TestFiles, attr_name)
        if skip_encrypted and member in encrypted:
            continue
        
        yield member


def test_paths():
    
    dirs = (TestDir, SourceTree, ResourceDir, OutputDir)
    for dirpath in dirs:
        print(dirpath)
        assert os.path.isdir(dirpath)
    
    for filepath in iterate_testfiles(False):
        print(filepath)
        assert os.path.isfile(filepath)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import pytest
import pypdfium2 as pdfium


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (0, 0),
        (90, 1),
        (180, 2),
        (270, 3),
    ]
)
def test_translate_rotation(test_input, expected):
    translated = pdfium.translate_rotation(test_input)
    assert translated == expected


@pytest.mark.parametrize(
    "values, expected",
    [
        ((255, 255, 255, 255), (0xFFFFFFFF, False)),
        ((255, 255, 255), (0xFFFFFFFF, False)),
        ((0, 255, 255, 255), (0xFF00FFFF, False)),
        ((255, 0, 255, 255), (0xFFFF00FF, False)),
        ((255, 255, 0, 255), (0xFFFFFF00, False)),
        ((255, 255, 255, 0), (0x00FFFFFF, True)),
    ]
)
def test_colour_to_hex(values, expected):
    assert pdfium.colour_as_hex(*values) == expected



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from enum import Enum


class OptimiseMode (Enum):
    """ How to optimise page rendering """
    
    none = 0          #: Do not use any optimisations
    lcd_display = 1   #: Optimise for LCD displays
    printing = 2      #: Optimise for printing


class ViewMode (Enum):
    """
    Modes that define how target coordinates of a bookmark should be interpreted.
    Refer to the PDF 1.6 reference manual, section 8.2 for more information.
    """
    
    Unknown = 0
    XYZ     = 1
    Fit     = 2
    FitH    = 3
    FitV    = 4
    FitR    = 5
    FitB    = 6
    FitBH   = 7
    FitBV   = 8



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parser import ArgParser


def _merge_pdfs(input_paths):
    
    dest_doc = pdfium.FPDF_CreateNewDocument()
    
    for in_path in reversed(input_paths):
        with pdfium.PdfContext(in_path) as src_doc:
            page_count = pdfium.FPDF_GetPageCount(src_doc)
            page_indices = (ctypes.c_int * page_count)(*[i for i in range(page_count)])
            pdfium.FPDF_ImportPagesByIndex(dest_doc, src_doc, page_indices, page_count, 0)
    
    return dest_doc


def parse_args(argv, prog, desc):
    
    parser = ArgParser(
        prog = prog,
        description = desc,
    )
    
    parser.add_argument(
        'inputs',
        nargs = '+',
        help = "A sequence of PDF files to concatenate",
    )
    parser.add_argument(
        '--output', '-o',
        required = True,
        help = "Target path for the output document",
    )
    
    return parser.parse_args(argv)


def main(argv, prog, desc):
    
    args = parse_args(argv, prog, desc)
    merged_doc = _merge_pdfs(args.inputs)
    
    with open(args.output, 'wb') as file_handle:
        pdfium.save_pdf(merged_doc, file_handle)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

V_MAJOR = 1
V_MINOR = 3
V_PATCH = 0
V_BETA = None

V_PYPDFIUM2 = "{}.{}.{}".format(V_MAJOR, V_MINOR, V_PATCH)  #: pypdfium2 version string
if V_BETA is not None:
    V_PYPDFIUM2 += "b{}".format(V_BETA)

V_LIBPDFIUM = 4983  #: PDFium library version integer (git tag)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import ctypes
from os.path import abspath
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.error_handler import handle_pdfium_error


def is_buffer(obj):
    """
    Check whether an object is a byte buffer that implements ``seek()``, ``tell()``,
    and ``readinto()``.
    
    Returns:
        :data:`True` if the object implements all required methods, :data:`False` otherwise.
    """
    
    if all( callable(getattr(obj, a, None)) for a in ('seek', 'tell', 'readinto') ):
        return True
    else:
        return False


class _reader_class:
    """
    Class that implements the callback for ``FPDF_FILEACCESS.m_GetBlock()``, to incrementally read file data from a buffer.
    """
    
    def __init__(self, buffer):
        self.buffer = buffer
    
    def __call__(self, param, position, p_buf, size):
        c_buf = (ctypes.c_char * size).from_address( ctypes.addressof(p_buf.contents) )
        self.buffer.seek(position)
        self.buffer.readinto(c_buf)
        return 1


class LoaderDataHolder:
    """
    Class to store data associated to an ``FPDF_DOCUMENT`` that was opened using ``FPDF_LoadCustomDocument()``.
    
    Parameters:
        file_handle:
            File buffer that implements the ``close()`` method.
        reader_instance:
            File access callable that must remain available for as long as the PDF is worked with.
    """
    
    def __init__(
            self,
            file_handle = None,
            reader_instance = None,
        ):
        self.file_handle = file_handle
        self.reader_instance = reader_instance
    
    def close(self):
        
        # access the reader variable again to make sure that even a heavily optimising interpreter would not prematurely delete the object
        id(self.reader_instance)
        
        if self.file_handle is not None:
            self.file_handle.close()


def open_pdf_buffer(buffer, password=None):
    """
    Open a PDF document incrementally from a byte buffer using ``FPDF_LoadCustomDocument()`` and a ctypes callback function.
    
    Parameters:
        buffer (typing.BinaryIO):
            A byte buffer as defined in :func:`.is_buffer`.
        password (str | bytes | None):
            A password to unlock the document, if encrypted.
    
    Returns:
        ``Tuple[pdfium.FPDF_DOCUMENT, LoaderDataHolder]``
    
    See also :func:`.open_pdf_auto`. **The same warnings apply!**
    """
    
    if not is_buffer(buffer):
        raise ValueError("Buffer must implement the methods seek(), tell(), and readinto().")
    
    buffer.seek(0, 2)
    file_len = buffer.tell()
    buffer.seek(0)
    
    FuncType = ctypes.CFUNCTYPE(
        # restype
        ctypes.c_int,
        # argtypes
        ctypes.POINTER(None),
        ctypes.c_ulong,
        ctypes.POINTER(ctypes.c_ubyte),
        ctypes.c_ulong,
    )
    
    fileaccess = pdfium.FPDF_FILEACCESS()
    fileaccess.m_FileLen = file_len
    fileaccess.m_GetBlock = FuncType( _reader_class(buffer) )
    
    pdf = pdfium.FPDF_LoadCustomDocument(ctypes.byref(fileaccess), password)
    ld_data = LoaderDataHolder(buffer, fileaccess.m_GetBlock)
    
    if pdfium.FPDF_GetPageCount(pdf) < 1:
        handle_pdfium_error(False)
    
    return pdf, ld_data


def open_pdf_native(filepath, password=None):
    """
    Open a PDF document from a file path, managing all file access natively in Python using :func:`.open_pdf_buffer`, without having to load the whole file into memory at once. This provides more independence from file access in PDFium.
    
    Parameters:
        filepath (str):
            File path to a PDF document.
        password (str | bytes | None):
            A password to unlock the document, if encrypted.
    
    Returns:
        ``Tuple[pdfium.FPDF_DOCUMENT, LoaderDataHolder]``
    
    See also :func:`.open_pdf_auto`. **The same warnings apply!**
    """
    
    file_handle = open(abspath(filepath), 'rb')
    return open_pdf_buffer(file_handle, password)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import concurrent.futures
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers import page_renderer
from pypdfium2._helpers.opener import PdfContext
from pypdfium2._helpers.constants import OptimiseMode


def _process_page(render_meth, input_obj, index, password, scale, rotation, colour, annotations, greyscale, optimise_mode):
    
    with PdfContext(input_obj, password) as pdf:
        result = render_meth(
            pdf, index,
            scale = scale,
            rotation = rotation,
            colour = colour,
            annotations = annotations,
            greyscale = greyscale,
            optimise_mode = optimise_mode,
        )
    
    return index, result


def _invoke_process_page(args):
    return _process_page(*args)


def render_pdf_base(
        render_meth,
        input_obj,
        page_indices = None,
        password = None,
        n_processes = os.cpu_count(),
        scale = 1,
        rotation = 0,
        colour = (255, 255, 255, 255),
        annotations = True,
        greyscale = False,
        optimise_mode = OptimiseMode.none,
    ):
    """
    Rasterise multiple pages of a PDF using an arbitrary page rendering method.
    Base function for :func:`.render_pdf_tobytes` and :func:`.render_pdf_topil`.
    
    Parameters:
        input_obj (str | bytes | typing.BinaryIO):
            The PDF document to render. It may be given as file path, bytes, or byte buffer.
        page_indices (typing.Sequence[int]):
            A list of zero-based page indices to render.
    
    The other parameters are the same as for :func:`.render_page_base`.
    """
    
    with PdfContext(input_obj, password) as pdf:
        n_pages = pdfium.FPDF_GetPageCount(pdf)
    n_digits = len(str(n_pages))
    
    if page_indices is None or len(page_indices) == 0:
        page_indices = [i for i in range(n_pages)]
    if not all(0 <= i < n_pages for i in page_indices):
        raise ValueError("Out of range page index detected.")
    
    args = [(render_meth, input_obj, i, password, scale, rotation, colour, annotations, greyscale, optimise_mode) for i in page_indices]
    
    with concurrent.futures.ProcessPoolExecutor(n_processes) as pool:
        for index, image in pool.map(_invoke_process_page, args):
            suffix = str(index+1).zfill(n_digits)
            yield image, suffix


def render_pdf_tobytes(*args, **kws):
    """
    Render multiple pages of a PDF to bytes. See :func:`.render_pdf_base` and :func:`.render_page_tobytes`.
    
    Yields:
        :class:`tuple`, :class:`str`
    """
    yield from render_pdf_base(page_renderer.render_page_tobytes, *args, **kws)


def render_pdf_topil(*args, **kws):
    """
    Render multiple pages of a PDF to :mod:`PIL` images. See :func:`.render_pdf_base` and :func:`.render_page_topil`.
    
    Yields:
        :class:`PIL.Image.Image`, :class:`str`
    """
    yield from render_pdf_base(page_renderer.render_page_topil, *args, **kws)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.constants import ViewMode


def translate_viewmode(viewmode):
    """
    Convert a PDFium view mode integer to an attribute of the :class:`.ViewMode` enum.
    
    Parameters:
        viewmode (int): PDFium view mode integer.
    
    Returns:
        :class:`.ViewMode`
    """
    
    if viewmode == pdfium.PDFDEST_VIEW_UNKNOWN_MODE:
        return ViewMode.Unknown
    elif viewmode == pdfium.PDFDEST_VIEW_XYZ:
        return ViewMode.XYZ
    elif viewmode == pdfium.PDFDEST_VIEW_FIT:
        return ViewMode.Fit
    elif viewmode == pdfium.PDFDEST_VIEW_FITH:
        return ViewMode.FitH
    elif viewmode == pdfium.PDFDEST_VIEW_FITV:
        return ViewMode.FitV
    elif viewmode == pdfium.PDFDEST_VIEW_FITR:
        return ViewMode.FitR
    elif viewmode == pdfium.PDFDEST_VIEW_FITB:
        return ViewMode.FitB
    elif viewmode == pdfium.PDFDEST_VIEW_FITBH:
        return ViewMode.FitBH
    elif viewmode == pdfium.PDFDEST_VIEW_FITBV:
        return ViewMode.FitBV
    else:
        raise ValueError("Unknown PDFium viewmode value {}".format(viewmode))


def translate_rotation(rotation):
    """
    Convert a rotation value in degrees to a PDFium rotation constant.
    
    Parameters:
        rotation (int): Rotation value in degrees (0, 90, 180, 270).
    
    Returns:
        :class:`int` – A PDFium rotation constant (0, 1, 2, 3).
    """
    
    if rotation == 0:
        return 0
    elif rotation == 90:
        return 1
    elif rotation == 180:
        return 2
    elif rotation == 270:
        return 3
    else:
        raise ValueError("Invalid rotation {}".format(rotation))


def _hex_digits(c):
    
    hxc = hex(c)[2:]
    if len(hxc) == 1:
        hxc = "0" + hxc
    
    return hxc
    

def colour_as_hex(r, g, b, a=255):
    """
    Convert a colour given as values of red, green, blue, and alpha ranging from 0 to 255 to a single integer in 32-bit ARGB format.
    
    Returns:
        :class:`int`, :class:`bool` – The colour integer, and a logical value that is :data:`True` if an alpha channel is needed, or :data:`False` if it is not needed.
    """
    
    use_alpha = True
    if a == 255:
        use_alpha = False
    
    colours = (a, r, g, b)
    for c in colours:
        assert 0 <= c <= 255
    
    hxc_str = "0x"
    for c in colours:
        hxc_str += _hex_digits(c)
    
    hxc_int = int(hxc_str, 0)
    
    return hxc_int, use_alpha



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pypdfium2 as pdfium
from .conftest import TestFiles


def _load_pages(pdf):
    n_pages = pdfium.FPDF_GetPageCount(pdf)
    return [pdfium.FPDF_LoadPage(pdf, i) for i in range(n_pages)]

def _close_pages(pages):
    for page in pages:
        pdfium.FPDF_ClosePage(page)


def _assert_boxes_eq(box, expected):
    #print(box)
    assert type(box) is type(expected)
    assert tuple( [round(val, 4) for val in box] ) == expected


def _check_boxes(pages, mediaboxes, cropboxes):
    
    assert len(pages) == len(mediaboxes) == len(cropboxes)
    
    for page, exp_mediabox in zip(pages, mediaboxes):
        _assert_boxes_eq(pdfium.get_mediabox(page), exp_mediabox)
        
    for page, exp_cropbox in zip(pages, cropboxes):
        _assert_boxes_eq(pdfium.get_cropbox(page), exp_cropbox)
    
    # todo: add test files that actually contain BleedBox, TrimBox, and ArtBox
    
    for page, exp_bleedbox in zip(pages, cropboxes):
        _assert_boxes_eq(pdfium.get_bleedbox(page), exp_bleedbox)
    
    for page, exp_trimbox in zip(pages, cropboxes):
        _assert_boxes_eq(pdfium.get_trimbox(page), exp_trimbox)
    
    for page, exp_artbox in zip(pages, cropboxes):
        _assert_boxes_eq(pdfium.get_artbox(page), exp_artbox)


def test_boxes_normal():
    
    with pdfium.PdfContext(TestFiles.multipage) as pdf:
        
        pages = _load_pages(pdf)
        boxes = (
            (0, 0, 595.2756, 841.8897),
            (0, 0, 595.2756, 419.5275),
            (0, 0, 297.6378, 419.5275),
        )
        
        _check_boxes(pages, boxes, boxes)
        _close_pages(pages)


def test_mediabox_fallback():
    
    with pdfium.PdfContext(TestFiles.mediabox_missing) as pdf:
        
        pages = _load_pages(pdf)
        boxes = (
            (0, 0, 612, 792),
            (0, 0, 612, 792),
        )
        
        _check_boxes(pages, boxes, boxes)
        _close_pages(pages)


def test_cropbox_different():
    
    with pdfium.PdfContext(TestFiles.cropbox) as pdf:
        
        pages = _load_pages(pdf)
        
        mediaboxes = [(0, 0, 612, 792) for i in range(20)]
        for i in range(12, 16):
            mediaboxes[i] = (0, 0, 419.52, 595.32)
        
        cropboxes = [(53, 35, 559, 757) for i in range(20)]
        for i in range(12, 16):
            cropboxes[i] = (48, 86, 371.52, 509.32)
        
        _check_boxes(pages, mediaboxes, cropboxes)
        _close_pages(pages)



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
from os.path import (
    join,
    basename,
)
import shutil
import sysconfig
import setuptools
from glob import glob
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

from platform_setup.packaging_base import (
    Libnames,
    DataTree,
    SourceTree,
    ModuleDir,
    PlatformNames,
    extract_version,
)


def _clean():
    
    build_cache = join(SourceTree,'build')
    if os.path.exists(build_cache):
        shutil.rmtree(build_cache)
    
    delete_files = [join(ModuleDir, n) for n in (*Libnames, '_pypdfium.py')]
    
    for file in delete_files:
        if os.path.exists(file):
            os.remove(file)


def _copy_bindings(pl_name):
    
    # non-recursively collect all objects from the platform directory
    for src_path in glob(join(DataTree, pl_name, '*')):
        
        # copy platform-specific files into the sources, excluding possible directories
        if os.path.isfile(src_path):
            dest_path = join(ModuleDir, basename(src_path))
            shutil.copy(src_path, dest_path)


def _get_linux_tag(arch):
    return 'manylinux_2_17_{}.manylinux2014_{}'.format(arch, arch)


def _get_mac_tag(arch, *versions):
    
    assert len(versions) > 0
    
    template = 'macosx_{}_{}'
    
    tag = ''
    sep = ''
    for v in versions:
        tag += sep + template.format(v, arch)
        sep = '.'
    
    return tag


def _get_tag(pl_name):
    if pl_name == PlatformNames.darwin_x64:
        return _get_mac_tag('x86_64', '10_11', '11_0', '12_0')
    elif pl_name == PlatformNames.darwin_arm64:
        return _get_mac_tag('arm64', '11_0', '12_0')
    elif pl_name == PlatformNames.linux_x64:
        return _get_linux_tag('x86_64')
    elif pl_name == PlatformNames.linux_x86:
        return _get_linux_tag('i686')
    elif pl_name == PlatformNames.linux_arm64:
        return _get_linux_tag('aarch64')
    elif pl_name == PlatformNames.linux_arm32:
        return _get_linux_tag('armv7l')
    elif pl_name == PlatformNames.windows_x64:
        return 'win_amd64'
    elif pl_name == PlatformNames.windows_arm64:
        return 'win_arm64'
    elif pl_name == PlatformNames.windows_x86:
        return 'win32'
    elif pl_name == PlatformNames.sourcebuild:
        tag = sysconfig.get_platform()
        for char in ('-', '.'):
            tag = tag.replace(char, '_')
        return tag
    else:
        raise ValueError( "Unknown platform directory {}".format(pl_name) )


def _get_bdist(pl_name):
    
    class bdist (_bdist_wheel):
        
        def finalize_options(self, *args, **kws):
            _bdist_wheel.finalize_options(self, *args, **kws)
            self.plat_name_supplied = True
        
        def get_tag(self, *args, **kws):
            return 'py3', 'none', _get_tag(pl_name)
    
    return bdist


SetupKws = dict(
    version = extract_version('V_PYPDFIUM2'),
)


def mkwheel(pl_name):
    
    _clean()
    _copy_bindings(pl_name)
    
    setuptools.setup(
        package_data = {'': Libnames},
        cmdclass = {'bdist_wheel': _get_bdist(pl_name)},
        **SetupKws,
    )
    
    _clean()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: CC-BY-4.0

import ctypes
import argparse
import pypdfium2 as pdfium


def _merge_pdfs(input_paths):
    
    dest_doc = pdfium.FPDF_CreateNewDocument()
    
    for in_path in reversed(input_paths):
        with pdfium.PdfContext(in_path) as src_doc:
            page_count = pdfium.FPDF_GetPageCount(src_doc)
            page_indices = (ctypes.c_int * page_count)(*[i for i in range(page_count)])
            pdfium.FPDF_ImportPagesByIndex(dest_doc, src_doc, page_indices, page_count, 0)
    
    return dest_doc


def parse_args():
    parser = argparse.ArgumentParser(
        description = "Merge PDF files with pypdfium2.",
    )
    parser.add_argument(
        'input_paths',
        nargs = '+',
    )
    parser.add_argument(
        '--output-path', '-o',
        required = True,
    )
    return parser.parse_args()


def main(input_paths, output_path):
    merged_doc = _merge_pdfs(input_paths)
    with open(output_path, 'wb') as file_handle:
        pdfium.save_pdf(merged_doc, file_handle)
    

if __name__ == '__main__':
    args = parse_args()
    main(
        input_paths = args.input_paths,
        output_path = args.output_path,
    )



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Download the PDFium binaries and generate ctypes bindings

import os
from os.path import join
import sys
import shutil
import tarfile
import argparse
import traceback
import subprocess
import importlib.util
from urllib import request
from concurrent.futures import ThreadPoolExecutor


if __name__ == '__main__': sys.modules['platform_setup'] = importlib.util.module_from_spec( importlib.util.spec_from_file_location('platform_setup', join(os.path.dirname(os.path.abspath(__file__)), '__init__.py')) )

from platform_setup.packaging_base import (
    DataTree,
    VersionFile,
    PlatformNames,
    run_cmd,
    extract_version,
    call_ctypesgen,
)


ReleaseRepo = "https://github.com/bblanchon/pdfium-binaries"
ReleaseURL = ReleaseRepo + "/releases/download/chromium%2F"
ReleaseExtension = "tgz"
ReleaseNames = {
    PlatformNames.darwin_x64    : 'pdfium-mac-x64',
    PlatformNames.darwin_arm64  : 'pdfium-mac-arm64',
    PlatformNames.linux_x64     : 'pdfium-linux-x64',
    PlatformNames.linux_x86     : 'pdfium-linux-x86',
    PlatformNames.linux_arm64   : 'pdfium-linux-arm64',
    PlatformNames.linux_arm32   : 'pdfium-linux-arm',
    PlatformNames.windows_x64   : 'pdfium-win-x64',
    PlatformNames.windows_x86   : 'pdfium-win-x86',
    PlatformNames.windows_arm64 : 'pdfium-win-arm64',
}


def _set_versions(*versions_list):
    
    with open(VersionFile, 'r') as fh:
        content = fh.read()
    
    for variable, current_ver, new_ver in versions_list:
        
        template = "{} = {}"
        previous = template.format(variable, current_ver)
        updated = template.format(variable, new_ver)
        
        print( "'{}' -> '{}'".format(previous, updated) )
        assert content.count(previous) == 1
        content = content.replace(previous, updated)
    
    with open(VersionFile, 'w') as fh:
        fh.write(content)


def get_latest_version():
    
    git_ls = run_cmd(['git', 'ls-remote', '{}.git'.format(ReleaseRepo)], cwd=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    
    git_ls = git_ls.stdout.decode('UTF-8')
    tag = git_ls.split('\t')[-1].replace('\n', '')
    version = int(tag.split('/')[-1])
    
    return version


def handle_versions(latest_version):
    
    v_minor = extract_version('V_MINOR')
    v_libpdfium = extract_version('V_LIBPDFIUM')
    
    if v_libpdfium < latest_version:
        print("New PDFium build")
        _set_versions(
            ('V_MINOR', v_minor, v_minor+1),
            ('V_LIBPDFIUM', v_libpdfium, latest_version),
        )
    
    else:
        print("No new PDFium build - will re-create bindings without incrementing version")


def clear_data(download_files):
    for pl_dir in download_files:
        if os.path.isdir(pl_dir):
            shutil.rmtree(pl_dir)


def _get_package(args):
    
    dirpath, file_url, file_path = args
    print("Downloading {} -> {}".format(file_url, file_path))
    
    try:
        request.urlretrieve(file_url, file_path)
    except Exception:
        traceback.print_exc()
        return None
    
    return dirpath, file_path


def download_releases(latest_version, download_files):
    
    base_url = "{}{}/".format(ReleaseURL, latest_version)
    args = []
    
    for dirpath, arcname in download_files.items():
        
        filename = "{}.{}".format(arcname, ReleaseExtension)
        file_url = base_url + filename
        
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        
        file_path = join(dirpath, filename)
        args.append( (dirpath, file_url, file_path) )
    
    archives = {}
    
    with ThreadPoolExecutor() as pool:
        
        for output in pool.map(_get_package, args):
            
            if output is not None:
                dirpath, file_path = output
                archives[dirpath] = file_path
    
    return archives


def unpack_archives(archives):
    
    for file in archives.values():
        
        extraction_path = join(os.path.dirname(file), 'build_tar')
        
        if ReleaseExtension == 'tgz':
            arc_opener = tarfile.open
        else:
            raise ValueError("Unknown archive extension {}".format(ReleaseExtension))
        
        with arc_opener(file) as archive:
            archive.extractall(extraction_path)
        
        os.remove(file)


def generate_bindings(archives): 
    
    for platform_dir in archives.keys():
        
        build_dir = join(platform_dir,'build_tar')
        bin_dir = join(build_dir,'lib')
        dirname = os.path.basename(platform_dir)
        
        if dirname.startswith('windows'):
            target_name = 'pdfium.dll'
            bin_dir = join(build_dir,'bin')
        elif dirname.startswith('darwin'):
            target_name = 'pdfium.dylib'
        elif dirname.startswith('linux'):
            target_name = 'pdfium'
        else:
            raise ValueError("Unknown platform directory name '{}'".format(dirname))
        
        items = os.listdir(bin_dir)
        assert len(items) == 1
        
        shutil.move(join(bin_dir, items[0]), join(platform_dir, target_name))
        
        call_ctypesgen(platform_dir, join(build_dir,'include'))
        shutil.rmtree(build_dir)


def get_download_files(platforms):
    
    avail_keys = [k for k in ReleaseNames.keys()]
    if platforms is None:
        platforms = avail_keys
    
    download_files = {}
    
    for pl_name in platforms:
        
        if pl_name in ReleaseNames:
            download_files[ join(DataTree, pl_name) ] = ReleaseNames[pl_name]
        else:
            raise ValueError(
                "Unknown platform name '{}'. Available keys are {}.".format(pl_name, avail_keys)
            )
    
    return download_files


def main(platforms):
    
    download_files = get_download_files(platforms)
    
    latest_version = get_latest_version()
    handle_versions(latest_version)
    clear_data(download_files)
    
    archives = download_releases(latest_version, download_files)
    unpack_archives(archives)
    generate_bindings(archives)


def parse_args(argv):
    parser = argparse.ArgumentParser(
        description = "Download pre-built PDFium packages and generate bindings",
    )
    parser.add_argument(
        '--platforms', '-p',
        metavar = 'P',
        nargs = '*',
    )
    return parser.parse_args(argv)


def run_cli(argv=sys.argv[1:]):
    args = parse_args(argv)
    return main(args.platforms)


if __name__ == '__main__':
    run_cli()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: CC-BY-4.0

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import time

project = 'pypdfium2'
copyright = '{} pypdfium2-team'.format( time.strftime('%Y') )
author = 'pypdfium2-team'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinxcontrib.programoutput',
    'myst_parser',
]

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

autodoc_preserve_defaults = True

autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'member-order': 'bysource',
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'PIL': ('https://pillow.readthedocs.io/en/stable/', None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parser import ArgParser


def parse_args(argv, prog, desc):
    parser = ArgParser(
        prog = prog,
        description = desc,
    )
    parser.add_argument(
        'input',
        help = "PDF document of which to print the outline",
    )
    parser.add_argument(
        '--max-depth',
        type = int,
        default = 15,
        help = "Maximum recursion depth to consider when parsing the table of contents",
    )
    return parser.parse_args(argv)

def print_toc(toc):
    for item in toc:
        print(
            '    ' * item.level +
            '{} -> {}  # {} {}'.format(
                item.title,
                item.page_index + 1,
                item.view_mode,
                item.view_pos,
            )
        )


def main(argv, prog, desc):
    args = parse_args(argv, prog, desc)
    doc = pdfium.PdfDocument(args.input)
    print_toc( doc.get_toc() )
    doc.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
from PIL import Image
from os.path import join
import pypdfium2 as pdfium
from .conftest import TestFiles, OutputDir


def _check_render_normal(pdf):
    
    pil_image = pdfium.render_page_topil(pdf, 0)
    
    assert pil_image.mode == 'RGB'
    assert pil_image.size == (595, 842)
    assert pil_image.getpixel( (0, 0) ) == (255, 255, 255)
    assert pil_image.getpixel( (150, 180) ) == (129, 212, 26)
    assert pil_image.getpixel( (150, 390) ) == (42, 96, 153)
    assert pil_image.getpixel( (150, 570) ) == (128, 0, 128)
    
    pil_image.close()


def test_render_page_filepath():
    with pdfium.PdfContext(TestFiles.render) as pdf:
        _check_render_normal(pdf)


def test_render_page_bytes():
    with open(TestFiles.render, 'rb') as fh:
        data = fh.read()
    with pdfium.PdfContext(data) as pdf:
        _check_render_normal(pdf)


def test_render_nonascii():
    with pdfium.PdfContext(TestFiles.nonascii) as pdf:
        pil_image = pdfium.render_page_topil(pdf, 0)
        pil_image.save( join(OutputDir,'render_nonascii.png') )


def _check_render_encrypted(file_or_data):
    
    with pdfium.PdfContext(file_or_data, 'test_user') as pdf:
        pil_image_a = pdfium.render_page_topil(pdf, 0)
    assert pil_image_a.mode == 'RGB'
    assert pil_image_a.size == (596, 842)
    
    with pdfium.PdfContext(file_or_data, 'test_owner') as pdf:
        pil_image_b = pdfium.render_page_topil(pdf, 0)
    assert pil_image_b.mode == 'RGB'
    assert pil_image_b.size == (596, 842)
    
    assert pil_image_a == pil_image_b
    
    pil_image_a.close()
    pil_image_b.close()


def test_render_page_encypted_file():
    _check_render_encrypted(TestFiles.encrypted)

def test_render_page_encrypted_bytes():
    with open(TestFiles.encrypted, 'rb') as fh:
        data = fh.read()
    _check_render_encrypted(data)


def test_render_page_alpha():
    
    with pdfium.PdfContext(TestFiles.render) as pdf:
        pil_image = pdfium.render_page_topil(
            pdf, 0,
            colour = None,
        )
    
    assert pil_image.mode == 'RGBA'
    assert pil_image.size == (595, 842)
    assert pil_image.getpixel( (0, 0) ) == (0, 0, 0, 0)
    assert pil_image.getpixel( (62, 66) ) == (0, 0, 0, 186)
    assert pil_image.getpixel( (150, 180) ) == (129, 212, 26, 255)
    assert pil_image.getpixel( (150, 390) ) == (42, 96, 153, 255)
    assert pil_image.getpixel( (150, 570) ) == (128, 0, 128, 255)
    
    pil_image.close()


def test_render_page_rotation():
    
    with pdfium.PdfContext(TestFiles.render) as pdf:
        
        image_0 = pdfium.render_page_topil(
            pdf, 0,
            rotation = 0
        )
        image_0.save(join(OutputDir,'rotate_0.png'))
        image_0.close()
        
        image_90 = pdfium.render_page_topil(
            pdf, 0,
            rotation = 90
        )
        image_90.save(join(OutputDir,'rotate_90.png'))
        image_90.close()
        
        image_180 = pdfium.render_page_topil(
            pdf, 0,
            rotation = 180
        )
        image_180.save(join(OutputDir,'rotate_180.png'))
        image_180.close()
        
        image_270 = pdfium.render_page_topil(
            pdf, 0,
            rotation = 270
        )
        image_270.save(join(OutputDir,'rotate_270.png'))
        image_270.close()


def test_render_pdf():
    
    with pdfium.PdfContext(TestFiles.multipage) as pdf:
        n_pages = pdfium.FPDF_GetPageCount(pdf)
        
    n_digits = len(str(n_pages))
    i = 0
    
    renderer = pdfium.render_pdf_topil(
        TestFiles.multipage,
        colour = (255, 255, 255),
    )
    for image, suffix in renderer:
        assert isinstance(image, Image.Image)
        assert suffix == str(i+1).zfill(n_digits)
        image.close()
        i += 1


def test_render_pdf_frombytes():
    
    with open(TestFiles.multipage, 'rb') as file_handle:
        file_bytes = file_handle.read()
    
    for image, suffix in pdfium.render_pdf_topil(file_bytes):
        assert isinstance(image, Image.Image)
        assert isinstance(suffix, str)
        assert image.mode == 'RGB'
        image.close()


def test_render_greyscale():
    
    with pdfium.PdfContext(TestFiles.render) as pdf:
        
        image_a = pdfium.render_page_topil(
            pdf, 0,
            greyscale = True,
        )
        image_a.save(join(OutputDir,'greyscale.png'))
        assert image_a.mode == 'L'
        image_a.close()
        
        image_b = pdfium.render_page_topil(
            pdf, 0,
            greyscale = True,
            colour = None,
        )
        assert image_b.mode == 'RGBA'
        image_b.save(join(OutputDir,'greyscale_alpha.png'))
        image_b.close()


@pytest.mark.parametrize(
    "colour",
    [
        (255, 255, 255, 255),
        (60, 70, 80, 100),
        (255, 255, 255),
        (0, 255, 255),
        (255, 0, 255),
        (255, 255, 0),
    ]
)
def test_render_bgcolour(colour):
    
    with pdfium.PdfContext(TestFiles.render) as pdf:
        pil_image = pdfium.render_page_topil(
            pdf, 0,
            colour = colour,
        )
    
    px_colour = colour
    if len(colour) == 4:
        if colour[3] == 255:
            px_colour = colour[:-1]
    
    bg_pixel = pil_image.getpixel( (0, 0) )
    assert bg_pixel == px_colour
    
    pil_image.close()


def _abstest_render_tobytes(img_info):
    data, cl_format, size = img_info
    assert isinstance(data, bytes)
    assert isinstance(cl_format, str)
    assert 1 <= len(cl_format) <= 4
    assert all (c in ('RGBLA') for c in cl_format)
    assert len(size) == 2
    assert all(isinstance(s, int) for s in size)
    assert len(data) == size[0] * size[1] * len(cl_format)


def test_render_page_tobytes():
    with pdfium.PdfContext(TestFiles.render) as pdf:
        _abstest_render_tobytes( pdfium.render_page_tobytes(pdf, 0) )


def test_render_pdf_tobytes():
    for img_info, num in pdfium.render_pdf_tobytes(TestFiles.multipage):
        _abstest_render_tobytes(img_info)
        assert num.isdigit()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: CC-BY-4.0 

import os
import sys
import pypdfium2 as pdfium


if __name__ == '__main__':
    if len(sys.argv) == 1:
        print("Usage: example.py somefile.pdf")
        sys.exit()
    
    filename = sys.argv[1]
    
    with pdfium.PdfContext(filename) as pdf:
        n_pages = pdfium.FPDF_GetPageCount(pdf)
    
    page_indices = [i for i in range(n_pages)]
    
    generator = pdfium.render_pdf_topil(
        filename,
        page_indices = page_indices,
        scale = 1,
        rotation = 0,
        colour = (255, 255, 255, 255),
        annotations = True,
        greyscale = False,
        optimise_mode = pdfium.OptimiseMode.none,
        n_processes = os.cpu_count(),
    )
    
    for image, suffix in generator:
        image.save("out_{}.png".format(suffix))
        image.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import logging
import pypdfium2 as pdfium
from pypdfium2._cli.toc import print_toc
from .conftest import TestFiles


def test_read_toc():
    
    with pdfium.PdfContext(TestFiles.bookmarks) as pdf:
        toc = pdfium.get_toc(pdf)
        print()
        print_toc(toc)


def test_read_toc_circular(caplog):
    with caplog.at_level(logging.CRITICAL):
        with pdfium.PdfContext(TestFiles.bookmarks_circular) as pdf:
            toc = pdfium.get_toc(pdf)
            print_toc(toc)
            assert "circular bookmark reference" in caplog.text



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause  

import io
from PIL import Image
from .conftest import (
    TestFiles,
    iterate_testfiles,
)
import pypdfium2 as pdfium
from pypdfium2._cli.toc import print_toc


def _pdfdoc_load(input_obj):
    doc = pdfium.PdfDocument(input_obj)
    page = pdfium.FPDF_LoadPage(doc.raw, 0)
    pdfium.FPDF_ClosePage(page)
    doc.close()


def test_pdfdoc_loadfiles():
    for filepath in iterate_testfiles():
        _pdfdoc_load(filepath)


def test_pdfdoc_renderpage():
    doc = pdfium.PdfDocument(TestFiles.render)
    image = doc.render_page_topil(0)
    assert isinstance(image, Image.Image)
    doc.close()


def test_pdfdoc_renderpdf():
    
    doc = pdfium.PdfDocument(TestFiles.multipage)
    
    i = 0
    for image, suffix in doc.render_pdf_topil():
        assert isinstance(image, Image.Image)
        assert isinstance(suffix, str)
        i+= 1
    assert i == 3
    
    doc.close()


def test_pdfdoc_save():
    doc = pdfium.PdfDocument(TestFiles.multipage)
    pdfium.FPDFPage_Delete(doc.raw, 0)
    buffer = io.BytesIO()
    doc.save(buffer)
    doc.close()
    assert buffer.tell() > 100000
    buffer.close()


def _compare_numarray(array, expected):
    assert type(array) is type(expected)
    assert len(array) == len(expected)
    for val_a, val_b in zip(array, expected):
        assert round(val_a) == val_b


def _compare_bookmark(
        bookmark,
        title,
        page_index,
        view_mode,
        view_pos,
    ):
    assert bookmark.title == title
    assert bookmark.page_index == page_index
    assert bookmark.view_mode == view_mode
    _compare_numarray(bookmark.view_pos, view_pos)


def test_pdfdoc_gettoc():
    
    doc = pdfium.PdfDocument(TestFiles.bookmarks)
    toc = doc.get_toc()
    
    # check first bookmark
    _compare_bookmark(
        next(toc),
        title = "One",
        page_index = 0,
        view_mode = pdfium.ViewMode.XYZ,
        view_pos = [89, 758, 0],
    )
    
    # check common values
    for bookmark in toc:
        assert isinstance(bookmark, pdfium.OutlineItem)
        assert bookmark.view_mode is pdfium.ViewMode.XYZ
        assert round(bookmark.view_pos[0]) == 89
    
    # check last bookmark
    _compare_bookmark(
        bookmark,
        title = "Three-B",
        page_index = 1,
        view_mode = pdfium.ViewMode.XYZ,
        view_pos = [89, 657, 0]
    )
    
    doc.close()


def test_pdfdoc_gettoc_maxdepth():
    
    doc = pdfium.PdfDocument(TestFiles.bookmarks)
    toc = doc.get_toc(max_depth=1)
    
    i = 0
    for bookmark in toc: i+= 1
    assert i == 3
    
    doc.close()


def test_pdfdoc_gettoc_print():
    doc = pdfium.PdfDocument(TestFiles.bookmarks)
    print_toc( doc.get_toc(max_depth=1) )
    doc.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import argparse


HelpFlags = ('help', '--help', '-h', '/h', '?', '/?')
VersionFlags = ('version', '--version', '-v')


class ArgParser (argparse.ArgumentParser):
    
    # Cutsomised argument parser
    
    def __init__(self, *args, **kws):
        super().__init__(*args, **kws)

    def parse_args(self, argv=None, namespace=None):
        
        if argv is None:
            argv = sys.argv[1:]
        
        if len(argv) < 1 or (argv[0].lower() in HelpFlags):
            self.print_help(sys.stderr)
            sys.exit()
        
        return super().parse_args(argv, namespace)


class _SubcommandItem:
    
    def __init__(
            self,
            names,
            method,
            help = "",
        ):
        if isinstance(names, str):
            self.names = (names, )
        else:
            self.names = names
        self.method = method
        self.help = help


class CliParser:
    
    def __init__(
            self,
            program,
            version,
            description,
            argv = sys.argv,
        ):
        self.program = program
        self.version = version
        self.description = description
        self.argv = argv
        self._subs = []
    
    
    def add_subcommand(self, *args, **kws):
        self._subs.append( _SubcommandItem(*args, **kws) )
    
    
    @staticmethod
    def _get_cmd_help(flags, message):
        
        help = ''
        
        sep = ''
        for flag in flags:
            help += sep + flag
            sep = ', '
        
        if len(flags) > 1:
            help = '({})'.format(help)
        
        help += '\n' + ' '*4 + message
        
        return help
    
    
    def _get_main_help(self):
        
        help = ''
        
        help += '{} {}'.format(self.program, self.version) + '\n'
        help += self.description + '\n'*2
        
        help += self._get_cmd_help(
            HelpFlags, "Show the main help an exit",
        ) + '\n'
        help += self._get_cmd_help(
            VersionFlags, "Show the versions of bindings and library",
        ) + '\n'
        
        sep = ''
        for sub in self._subs:
            help += sep + self._get_cmd_help(sub.names, sub.help)
            sep = '\n'
        
        return help
    
    
    def run(self):

        if len(self.argv) < 1 or self.argv[0] in HelpFlags:
            print(self._get_main_help())
            sys.exit()
        
        main_arg = self.argv[0]
        
        if main_arg in VersionFlags:
            print( "{} {}".format(self.program, self.version) )
            sys.exit()
        
        sc_found = False
        
        for sub in self._subs:
            if main_arg.lower() in [n.lower() for n in sub.names]:
                sub.method(
                    argv = self.argv[1:],
                    prog = "{} {}".format(self.program, main_arg),
                    desc = sub.help,
                )
                sc_found = True
                break
        
        if not sc_found:
            print("Error: Argument '{}' is not a valid subcommand".format(main_arg), file=sys.stderr)



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import sysconfig
import setuptools
import importlib.util
from os.path import (
    join,
    abspath,
    dirname,
    basename,
)


def include_platform_setup():
    
    mod_name = 'platform_setup'
    PlatSetupInit = join(dirname(abspath(__file__)), mod_name, '__init__.py')
    
    spec = importlib.util.spec_from_file_location(mod_name, PlatSetupInit)
    sys.modules[mod_name] = importlib.util.module_from_spec(spec)


def packaging_handler():
    
    target = os.environ.get('PYP_TARGET_PLATFORM', None)
    if target in (None, 'auto'):
        return True
    
    from platform_setup.setup_base import mkwheel, SetupKws
    from platform_setup.packaging_base import PlatformNames
    
    if target == 'sdist':
        setuptools.setup(**SetupKws)
    elif hasattr(PlatformNames, target):
        mkwheel( getattr(PlatformNames, target) )
    else:
        raise ValueError( "Invalid deployment target '{}'".format(target) )
    
    return False


def install_handler():
    
    from platform_setup import check_deps
    from platform_setup.packaging_base import SourceTree, PlatformNames
    
    StatusFile = join(SourceTree, 'platform_setup', '.presetup_done.txt')
    
    def check_presetup():
        if os.path.exists(StatusFile):
            return False
        else:
            with open(StatusFile, 'w') as fh:
                fh.write('')
            return True
    
    W_Presetup = check_presetup()
    if W_Presetup: check_deps.main()
    
    from platform_setup import build_pdfium, update_pdfium
    from platform_setup.setup_base import mkwheel
    
    class HostPlatform:
        
        def __init__(self):
            plat_name = sysconfig.get_platform().lower()
            for char in ('-', '.'):
                plat_name = plat_name.replace(char, '_')
            self.plat_name = plat_name
        
        def is_platform(self, start, end):
            if self.plat_name.startswith(start):
                if self.plat_name.endswith(end):
                    return True
            return False
    
    def _setup(pl_name):
        if W_Presetup: update_pdfium.main( [basename(pl_name)] )
        mkwheel(pl_name)
    
    host = HostPlatform()
    
    if host.is_platform('macosx', 'arm64'):
        _setup(PlatformNames.darwin_arm64)
    elif host.is_platform('macosx', 'x86_64'):
        _setup(PlatformNames.darwin_x64)
    elif host.is_platform('linux', 'armv7l'):
        _setup(PlatformNames.linux_arm32)
    elif host.is_platform('linux', 'aarch64'):
        _setup(PlatformNames.linux_arm64)
    elif host.is_platform('linux', 'x86_64'):
        _setup(PlatformNames.linux_x64)
    elif host.is_platform('linux', 'i686'):
        _setup(PlatformNames.linux_x86)
    elif host.is_platform('win', 'arm64'):
        _setup(PlatformNames.windows_arm64)
    elif host.is_platform('win', 'amd64'):
        _setup(PlatformNames.windows_x64)
    elif host.is_platform('win32', ''):
        _setup(PlatformNames.windows_x86)
    else:
        # Platform without pre-built binaries - try a regular sourcebuild
        if W_Presetup: build_pdfium.main()
        mkwheel(PlatformNames.sourcebuild)


def main():
    include_platform_setup()
    cont = packaging_handler()
    if cont: install_handler()

if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pkg_resources
import pypdfium2 as pdfium
from pypdfium2._version import (
    V_PYPDFIUM2,
    V_LIBPDFIUM,
)


def _get_pkg_version(pkgname):
    return pkg_resources.get_distribution(pkgname).version


def test_version_aliases():
    assert pdfium.__version__ == V_PYPDFIUM2 == _get_pkg_version('pypdfium2')
    assert pdfium.__pdfium_version__ == V_LIBPDFIUM



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
from pypdfium2._version import (
    V_PYPDFIUM2,
    V_LIBPDFIUM,
)
from pypdfium2._cli._parser import CliParser
from pypdfium2._cli import (
    renderer,
    toc,
    merger,
    tiler,
)


def main(argv=sys.argv[1:]):
    
    parser = CliParser(
        program = "pypdfium2",
        version = "{} (libpdfium {})".format(V_PYPDFIUM2, V_LIBPDFIUM),
        description = "Command line interface to the pypdfium2 Python library",
        argv = argv,
    )
    
    parser.add_subcommand(
        "render",
        method = renderer.main,
        help = "Rasteries pages of a PDF file",
    )
    parser.add_subcommand(
        "toc",
        method = toc.main,
        help = "Show the table of contents for a PDF document",
    )
    parser.add_subcommand(
        "merge",
        method = merger.main,
        help = "Concatenate PDF files",
    )
    parser.add_subcommand(
        "tile",
        method = tiler.main,
        help = "Perform page tiling (N-up compositing)",
    )
    
    parser.run()
